import 'package:minehunter/config/default_game_settings.dart';
import 'package:minehunter/models/game/cell.dart';
import 'package:minehunter/models/settings/settings_game.dart';
import 'package:minehunter/models/settings/settings_global.dart';
import 'package:minehunter/models/types.dart';
import 'package:minehunter/utils/tools.dart';

class Game {
  Game({
    // Settings
    required this.gameSettings,
    required this.globalSettings,

    // State
    this.isRunning = false,
    this.isStarted = false,
    this.isFinished = false,
    this.animationInProgress = false,

    // Base data
    required this.board,
    required this.sizeHorizontal,
    required this.sizeVertical,
    this.isBoardMined = false,

    // Game data
    this.minesCount = 0,
    this.reportMode = false,
    this.gameWin = false,
    this.gameFail = false,
  });

  // Settings
  final GameSettings gameSettings;
  final GlobalSettings globalSettings;

  // State
  bool isRunning;
  bool isStarted;
  bool isFinished;
  bool animationInProgress;

  // Base data
  Board board;
  final int sizeHorizontal;
  final int sizeVertical;
  bool isBoardMined;

  // Game data
  int minesCount;
  bool reportMode;
  bool gameWin;
  bool gameFail;

  factory Game.createEmpty() {
    return Game(
      // Settings
      gameSettings: GameSettings.createDefault(),
      globalSettings: GlobalSettings.createDefault(),
      // Base data
      board: [],
      sizeHorizontal: 0,
      sizeVertical: 0,
    );
  }

  factory Game.createNew({
    GameSettings? gameSettings,
    GlobalSettings? globalSettings,
  }) {
    final GameSettings newGameSettings = gameSettings ?? GameSettings.createDefault();
    final GlobalSettings newGlobalSettings = globalSettings ?? GlobalSettings.createDefault();

    final int sizeHorizontal = int.parse(newGameSettings.size.split('x')[0]);
    final int sizeVertical = int.parse(newGameSettings.size.split('x')[1]);

    // Create empty board (it will be mined after first hit)
    final Board board = [];
    for (int rowIndex = 0; rowIndex < sizeVertical; rowIndex++) {
      final List<Cell> row = [];
      for (int colIndex = 0; colIndex < sizeHorizontal; colIndex++) {
        row.add(Cell(false));
      }
      board.add(row);
    }

    final int minesCountRatio =
        DefaultGameSettings.minesCountRatios[newGameSettings.level] ?? 1;
    final int minesCount = ((sizeHorizontal * sizeVertical) * minesCountRatio / 100).round();
    printlog('Mines count: $minesCount');

    return Game(
      // Settings
      gameSettings: newGameSettings,
      globalSettings: newGlobalSettings,
      // State
      isRunning: true,
      // Base data
      board: board,
      sizeHorizontal: sizeHorizontal,
      sizeVertical: sizeVertical,
      // Game data
      minesCount: minesCount,
    );
  }

  void addMines({
    required int forbiddenRow,
    required int forbiddenCol,
  }) {
    // Shuffle cells to put random mines, except on currently selected one
    final List<List<int>> allowedCells = [];
    for (int row = 0; row < sizeVertical; row++) {
      for (int col = 0; col < sizeHorizontal; col++) {
        if (!((forbiddenRow == row) && (forbiddenCol == col))) {
          allowedCells.add([row, col]);
        }
      }
    }
    allowedCells.shuffle();

    // Put random mines on board
    for (int mineIndex = 0; mineIndex < minesCount; mineIndex++) {
      board[allowedCells[mineIndex][0]][allowedCells[mineIndex][1]].isMined = true;
    }

    // Compute all mines counts on cells
    for (int row = 0; row < sizeVertical; row++) {
      for (int col = 0; col < sizeHorizontal; col++) {
        board[row][col].minesCountAround = getMinesCountAround(row, col);
      }
    }

    printGrid();
  }

  int getMinesCountAround(int row, int col) {
    final int sizeHorizontal = board.length;
    final int sizeVertical = board[0].length;

    int minesCountAround = 0;
    for (int deltaRow = -1; deltaRow <= 1; deltaRow++) {
      for (int deltaCol = -1; deltaCol <= 1; deltaCol++) {
        if ((row + deltaRow >= 0 && row + deltaRow < sizeVertical) &&
            (col + deltaCol >= 0 && col + deltaCol < sizeHorizontal) &&
            (board[row + deltaRow][col + deltaCol].isMined)) {
          minesCountAround++;
        }
      }
    }

    return minesCountAround;
  }

  int countFlaggedCells() {
    int count = 0;

    final int sizeHorizontal = board.length;
    final int sizeVertical = board[0].length;
    for (int row = 0; row < sizeVertical; row++) {
      for (int col = 0; col < sizeHorizontal; col++) {
        if (board[row][col].isMarked == true) {
          count++;
        }
      }
    }

    return count;
  }

  bool get canBeResumed => isStarted && !isFinished;
  bool get gameWon => isRunning && isStarted && isFinished;

  void dump() {
    printlog('');
    printlog('## Current game dump:');
    printlog('');
    printlog('$Game:');
    printlog('  Settings');
    gameSettings.dump();
    globalSettings.dump();
    printlog('  State');
    printlog('    isRunning: $isRunning');
    printlog('    isStarted: $isStarted');
    printlog('    isFinished: $isFinished');
    printlog('    animationInProgress: $animationInProgress');
    printlog('  Base data');
    printGrid();
    printlog('    sizeHorizontal: $sizeHorizontal');
    printlog('    sizeVertical: $sizeVertical');
    printlog('    isBoardMined: $isBoardMined');
    printlog('  Game data');
    printlog('    minesCount: $minesCount');
    printlog('    reportMode: $reportMode');
    printlog('    gameWin: $gameWin');
    printlog('    gameFail: $gameFail');
    printlog('');
  }

  void printGrid() {
    const String isMined = 'X';
    const String isSafe = '.';

    const String mineFound = '#';
    const String wrongMarkedCell = '0';
    const String exploredSafeCell = '.';
    const String unkownState = ' ';

    printlog('');
    String line = '--';
    for (int i = 0; i < board[0].length; i++) {
      line += '-';
    }
    printlog('$line  $line');
    for (int rowIndex = 0; rowIndex < board.length; rowIndex++) {
      String currentLine = '';
      String solvedLine = '';
      for (int colIndex = 0; colIndex < board[rowIndex].length; colIndex++) {
        solvedLine += board[rowIndex][colIndex].isMined ? isMined : isSafe;

        String cellString = unkownState;
        if (board[rowIndex][colIndex].isExplored) {
          cellString = exploredSafeCell;
        }
        if (board[rowIndex][colIndex].isMarked) {
          if (board[rowIndex][colIndex].isMined) {
            cellString = mineFound;
          } else {
            cellString = wrongMarkedCell;
          }
        }
        currentLine += cellString;
      }
      printlog('|$currentLine|  |$solvedLine|');
    }
    printlog('$line  $line');
    printlog('');
  }

  @override
  String toString() {
    return '$Game(${toJson()})';
  }

  Map<String, dynamic>? toJson() {
    return <String, dynamic>{
      // Settings
      'gameSettings': gameSettings.toJson(),
      'globalSettings': globalSettings.toJson(),
      // State
      'isRunning': isRunning,
      'isStarted': isStarted,
      'isFinished': isFinished,
      'animationInProgress': animationInProgress,
      // Base data
      'board': board,
      'sizeHorizontal': sizeHorizontal,
      'sizeVertical': sizeVertical,
      'isBoardMined': isBoardMined,
      // Game data
      'minesCount': minesCount,
      'reportMode': reportMode,
      'gameWin': gameWin,
      'gameFail': gameFail,
    };
  }
}
