import 'dart:math';

import 'package:flutter/material.dart';

import 'package:minehunter/config/default_game_settings.dart';
import 'package:minehunter/models/settings/settings_game.dart';
import 'package:minehunter/models/settings/settings_global.dart';
import 'package:minehunter/utils/tools.dart';

class ParameterPainter extends CustomPainter {
  const ParameterPainter({
    required this.code,
    required this.value,
    required this.isSelected,
    required this.gameSettings,
    required this.globalSettings,
  });

  final String code;
  final String value;
  final bool isSelected;
  final GameSettings gameSettings;
  final GlobalSettings globalSettings;

  @override
  void paint(Canvas canvas, Size size) {
    // force square
    final double canvasSize = min(size.width, size.height);

    const Color borderColorEnabled = Colors.blue;
    const Color borderColorDisabled = Colors.white;

    // "enabled/disabled" border
    final paint = Paint();
    paint.style = PaintingStyle.stroke;
    paint.color = isSelected ? borderColorEnabled : borderColorDisabled;
    paint.strokeJoin = StrokeJoin.round;
    paint.strokeWidth = 10;
    canvas.drawRect(
        Rect.fromPoints(const Offset(0, 0), Offset(canvasSize, canvasSize)), paint);

    // content
    switch (code) {
      case DefaultGameSettings.parameterCodeLevel:
        paintLevelParameterItem(value, canvas, canvasSize);
        break;
      case DefaultGameSettings.parameterCodeSize:
        paintSizeParameterItem(value, canvas, canvasSize);
        break;
      default:
        printlog('Unknown parameter: $code/$value');
        paintUnknownParameterItem(value, canvas, canvasSize);
    }
  }

  @override
  bool shouldRepaint(CustomPainter oldDelegate) {
    return false;
  }

  // "unknown" parameter -> simple block with text
  void paintUnknownParameterItem(
    final String value,
    final Canvas canvas,
    final double size,
  ) {
    final paint = Paint();
    paint.strokeJoin = StrokeJoin.round;
    paint.strokeWidth = 3;

    paint.color = Colors.grey;
    paint.style = PaintingStyle.fill;
    canvas.drawRect(Rect.fromPoints(const Offset(0, 0), Offset(size, size)), paint);

    final textSpan = TextSpan(
      text: '?\n$value',
      style: const TextStyle(
        color: Colors.black,
        fontSize: 18,
        fontWeight: FontWeight.bold,
      ),
    );
    final textPainter = TextPainter(
      text: textSpan,
      textDirection: TextDirection.ltr,
      textAlign: TextAlign.center,
    );
    textPainter.layout();
    textPainter.paint(
      canvas,
      Offset(
        (size - textPainter.width) * 0.5,
        (size - textPainter.height) * 0.5,
      ),
    );
  }

  void paintLevelParameterItem(
    final String value,
    final Canvas canvas,
    final double size,
  ) {
    Color backgroundColor = Colors.grey;

    final List<dynamic> stars = [];

    switch (value) {
      case DefaultGameSettings.levelValueEasy:
        backgroundColor = Colors.green;
        stars.add([0.5, 0.5]);
        break;
      case DefaultGameSettings.levelValueMedium:
        backgroundColor = Colors.orange;
        stars.add([0.3, 0.5]);
        stars.add([0.7, 0.5]);
        break;
      case DefaultGameSettings.levelValueHard:
        backgroundColor = Colors.red;
        stars.add([0.3, 0.3]);
        stars.add([0.7, 0.3]);
        stars.add([0.5, 0.7]);
        break;
      case DefaultGameSettings.levelValueNightmare:
        backgroundColor = Colors.purple;
        stars.add([0.3, 0.3]);
        stars.add([0.7, 0.3]);
        stars.add([0.3, 0.7]);
        stars.add([0.7, 0.7]);
        break;
      default:
        printlog('Wrong value for level parameter value: $value');
    }

    final paint = Paint();
    paint.strokeJoin = StrokeJoin.round;
    paint.strokeWidth = 3 / 100 * size;

    // Colored background
    paint.color = backgroundColor;
    paint.style = PaintingStyle.fill;
    canvas.drawRect(Rect.fromPoints(const Offset(0, 0), Offset(size, size)), paint);

    // Stars
    final textSpan = TextSpan(
      text: '⭐',
      style: TextStyle(
        color: Colors.black,
        fontSize: size / 3,
        fontWeight: FontWeight.bold,
      ),
    );
    final textPainter = TextPainter(
      text: textSpan,
      textDirection: TextDirection.ltr,
      textAlign: TextAlign.center,
    );
    textPainter.layout();

    for (var center in stars) {
      textPainter.paint(
        canvas,
        Offset(
          size * center[0] - textPainter.width * 0.5,
          size * center[1] - textPainter.height * 0.5,
        ),
      );
    }
  }

  void paintSizeParameterItem(
    final String value,
    final Canvas canvas,
    final double size,
  ) {
    Color backgroundColor = Colors.grey;
    int gridWidth = 1;

    switch (value) {
      case DefaultGameSettings.sizeValueSmall:
        backgroundColor = Colors.green;
        gridWidth = 2;
        break;
      case DefaultGameSettings.sizeValueMedium:
        backgroundColor = Colors.orange;
        gridWidth = 3;
        break;
      case DefaultGameSettings.sizeValueLarge:
        backgroundColor = Colors.red;
        gridWidth = 4;
        break;
      default:
        printlog('Wrong value for boardSize parameter value: $value');
    }

    final paint = Paint();
    paint.strokeJoin = StrokeJoin.round;
    paint.strokeWidth = 3 / 100 * size;

    // Colored background
    paint.color = backgroundColor;
    paint.style = PaintingStyle.fill;
    canvas.drawRect(Rect.fromPoints(const Offset(0, 0), Offset(size, size)), paint);

    // Mini grid
    final squareColor = Colors.grey.shade200;
    final borderColor = Colors.grey.shade800;

    final double cellSize = size / 7;
    final double origin = (size - gridWidth * cellSize) / 2;

    for (int row = 0; row < gridWidth; row++) {
      for (int col = 0; col < gridWidth; col++) {
        final Offset topLeft = Offset(origin + col * cellSize, origin + row * cellSize);
        final Offset bottomRight = topLeft + Offset(cellSize, cellSize);

        paint.color = squareColor;
        paint.style = PaintingStyle.fill;
        canvas.drawRect(Rect.fromPoints(topLeft, bottomRight), paint);

        paint.color = borderColor;
        paint.style = PaintingStyle.stroke;
        canvas.drawRect(Rect.fromPoints(topLeft, bottomRight), paint);
      }
    }
  }
}
